import os
import typer
import uvicorn
from .restapi import app as app


def http():
    """
    Starts uvicorn serving the FastAPI app() defined in restapi.py on localhost port 8000.
    For local development only! Production deployment should use the containerized stack.
    """

    if is_docker():
        typer.echo(
            """
        You appear to be running proximatic-http from a Docker container. 
        Make sure this is not your production container. 
        Type Y to launch uvicorn on localhost:8000.
        """
        )
    # Launch a web browser on localhost open to the api endpoint.
    # @see https://typer.tiangolo.com/tutorial/launch/

    typer.echo("Launching web browser to api docs:")
    typer.launch("http://127.0.0.1:8000/docs")

    # Run unvicorn server bound to port 8000 on localhost.
    # @see https://fastapi.tiangolo.com/#example
    # @see https://www.uvicorn.org/deployment/

    uvicorn.run(
        "proximatic.runhttp:app",
        host="127.0.0.1",
        port=8000,
        reload=True,  # For development. Watches for file changes and reloads the server hot.
        log_level="debug",
    )


def is_docker():
    path = "/proc/self/cgroup"
    return (
        os.path.exists("/.dockerenv")
        or os.path.isfile(path)
        and any("docker" in line for line in open(path))
    )



import os
from fastapi import FastAPI, status
from fastapi.openapi.utils import get_openapi
from proximatic import Proximatic, __version__


app = FastAPI()

proximatic = Proximatic(fqdn=os.getenv("PROXIMATIC_FQDN"))


@app.get("/")
def read_root():
    """Endpoint returning basic information."""

    return {"Proximatic REST API": __version__}


@app.get("/provider/list")
def provider_list():
    """Endpoint that returns a list of configured providers."""

    response = proximatic.provider_list()

    return response

def proximatic_openapi():
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="Proximatic REST API",
        version=__version__,
        description="Proximatic OpenAPI schema",
        routes=app.routes,
    )
    openapi_schema["info"]["x-logo"] = {
        "url": "https://i.pinimg.com/474x/ea/e2/07/eae2071d69a1089f82ec60a5be6037c0.jpg"
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = proximatic_openapi



import typer
import pprint
from proximatic import Proximatic, __version__
from .utils import tabulate_resources as tabulate_resources

app = typer.Typer()

proximatic = Proximatic()

pp = pprint.PrettyPrinter(indent=4)


@app.callback()
def callback():
    """
    --------------\n
    Proximatic CLI\n
    Interactive command-line interface to Proximatic.\n
    --------------\n
    """


@app.command()
def config_show():
    typer.echo(pp.pprint(proximatic.system.dict()))


@app.command()
def provider_list(id: str = None):
    """Returns a list of configured providers."""
    response = proximatic.provider_list(id)
    if response.data:
        table = tabulate_resources(response)
        typer.echo(table)
    else:
        typer.echo(response.error)


@app.command()
def provider_create(id: str, server: str):
    response = proximatic.provider_create(id=id, server=server)
    if response.data:
        typer.echo(
            f"\nSuccessfully created {response.data[0].type} {response.data[0].id}.\n"
        )
        table = tabulate_resources(response)
        typer.echo(table)
    else:
        if response.error:
            typer.echo(pp.pprint(response.error[0].dict()))


@app.command()
def provider_delete(id: str):
    response = proximatic.provider_delete(id)
    typer.echo(pp.pprint(response.dict()))



import os
import yaml
import requests
from typing import List
from pathlib import Path
from .models import (
    SystemConfigModel,
    ProviderAttributesModel,
    ResponseModel,
    ResponseErrorModel,
    ResourceModel,
    RouterModel,
    ServiceModel,
    LoadBalancerModel,
    ProviderExportModel,
)


class Proximatic:
    """The proximatic core engine."""

    def __init__(self, yml_path: str = None, fqdn: str = None):
        """Bootstraps the Proximatic object with fqdn string and path to .yml files."""
        if yml_path:
            yml_path = yml_path
        elif os.getenv("PROXIMATIC_YML_PATH"):
            yml_path = os.getenv("PROXIMATIC_YML_PATH")
        else:
            yml_path = "./data"
        yml_path = Path(yml_path)
        if yml_path.exists() and yml_path.is_dir():
            if fqdn:
                fqdn = fqdn
            elif os.getenv("PROXIMATIC_FQDN"):
                fqdn = os.getenv("PROXIMATIC_FQDN")
            else:
                fqdn = "localhost"
            self.system = SystemConfigModel(yml_path=yml_path, fqdn=fqdn)
            self.load_config()
        else:
            raise Exception()

    def load_config(self) -> SystemConfigModel:
        """ETL function that reads all configuration in yml_path and loads resources into Proximatic()."""

        files = self.system.yml_path.glob("**/*.yml")
        for filename in files:
            with open(filename, "r") as yml_stream:
                # Load our yml file as dict.
                config = yaml.safe_load(yml_stream)
                # Extract the values into pydantic models.
                if "http" in config and "services" in config["http"]:
                    router_id = list(config["http"]["routers"].keys())[0]
                    # We want the service that the router references.
                    service_id = config["http"]["routers"][router_id]["service"]
                    # Create our load balancer instance from our model.
                    loadbalancer = LoadBalancerModel(
                        servers=config["http"]["services"][service_id]["loadBalancer"][
                            "servers"
                        ]
                    )
                    # Create our service instance from our model and attach our loadbalancer.
                    service = ServiceModel(id=service_id, loadBalancer=loadbalancer)
                    # Create our router instance from our model.
                    router = RouterModel(
                        id=router_id,
                        entryPoints=config["http"]["routers"][router_id]["entryPoints"],
                        rule=config["http"]["routers"][router_id]["rule"],
                        middlewares=config["http"]["routers"][router_id]["middlewares"],
                        service=service_id,
                    )
                    # Create middlewares declaration as dictionary:
                    middlewares = {}
                    for name, configuration in config["http"]["middlewares"].items():
                        middlewares[name] = configuration
                    # Create our provider attributes instance from our model
                    # and attach our data.
                    attributes = ProviderAttributesModel(
                        router=router,
                        service=service,
                        middlewares=middlewares,
                        endpoint=router.rule.split("`")[1],
                        server=service.loadBalancer.servers[0]["url"],
                    )

                    provider = ResourceModel(
                        id=router_id, type="provider", attributes=attributes
                    )

                    self.system.providers.append(provider)
        return self.system

    def set_fqdn(self, fqdn: str):
        self.system.fqdn = fqdn
        # @todo Decide if this should also rewrite the fqdn in all yml?

    def get_fqdn(self):
        return self.system.fqdn

    def provider_list(self, id: str = None) -> ResponseModel:
        """
        Returns a ResponseModel containing all Provider resources
        discovered in the active config.
        """
        response = ResponseModel()
        resources = []
        if id:
            providers = [provider for provider in self.system.providers if provider.id == id]
        else:
            providers = self.system.providers
        try:
            for provider in providers:
                resources.append(provider)
            response.data = resources
        except Exception as e:
            response.error = [ResponseErrorModel(id="changeme", detail=str(e))]
        return response

    def provider_fetch(self, id: str) -> List[ResourceModel]:
        self.load_config()
        return next((item for item in self.system.providers if item.id == id), None)

    def provider_export(self, provider: ResourceModel) -> ResponseModel:
        """File dump function that takes a provider resource model instance and writes it
        to a proxy configuration .yml file.
        """

        response = ResponseModel()
        if provider.type != "provider":
            error = ResponseErrorModel(
                id="Invalid type.", detail="We need a better error system."
            )
            response.error.append(error)
            return response
        export = ProviderExportModel()
        export.http["routers"][provider.id] = provider.attributes.router.dict(
            exclude={"id"}
        )
        export.http["services"][provider.id] = provider.attributes.service.dict()
        export.http["middlewares"] = provider.attributes.middlewares
        file_path = self.system.yml_path.joinpath(provider.id + ".yml")
        with open(file_path, "wt") as yml_stream:
            yaml.dump(export.dict(), yml_stream)
        return ResponseModel()

    def provider_create(self, id: str, server: str) -> ResponseModel:
        response = ResponseModel()
        fetched_provider = self.provider_fetch(id)
        if fetched_provider:
            response.error = [
                ResponseErrorModel(id="changeme", detail="Provider already exists.")
            ]
            return response
        # Check the URL for validity by visiting it and
        # expecting a 200 response code from its server.
        try:
            result = requests.get(server).status_code
            if result != 200:
                response.error = [
                    ResponseErrorModel(
                        id="changeme", detail="server URL not reachable."
                    )
                ]
                return response
        except Exception as e:
            response.error = [ResponseErrorModel(id="changeme", detail=str(e))]
            return response
        endpoint = f"{id}.{self.system.fqdn}"
        service = ServiceModel(
            loadBalancer=LoadBalancerModel(servers=[{"url": server}])
        )
        # Set the default headers middleware for this provider.
        middlewares = {
            f"{id}-headers": {
                "headers": {
                    "frameDeny": True,
                    "sslRedirect": True,
                    "browserXssFilter": True,
                    "contentTypeNosniff": True,
                    "forceSTSHeader": True,
                    "stsIncludeSubdomains": True,
                    "stsPreload": True,
                }
            }
        }
        router = RouterModel(
            id=id,
            rule=f"Host(`{endpoint}`)",
            service=id,
            middlewares = [f"{id}-headers"]
        )
        attributes = ProviderAttributesModel(
            router=router,
            middlewares=middlewares,
            service=service,
            endpoint=endpoint,
            server=server,
        )
        provider = ResourceModel(id=id, type="provider", attributes=attributes)
        self.provider_export(provider)
        self.load_config()
        fetched_provider = self.provider_fetch(id)
        if fetched_provider:
            response.data = [provider]
        else:
            response.error = [
                ResponseErrorModel(id="changeme", detail="Provider not created.")
            ]
        return response

    def provider_delete(self, id: str) -> ResponseModel:
        provider = self.provider_fetch(id)
        if provider and provider.id == id:
            files = self.system.yml_path.rglob(f"{provider.id}.yml")
            for path in files:
                path.unlink()
            return ResponseModel(meta={"result": "deleted"})
        else:
            return ResponseModel(meta={"result": "not found"})



from tabulate import tabulate
from .models import ResponseModel


def tabulate_resources(response: ResponseModel) -> str:
    """
    Takes a ResponseModel and returns all of its resources as a string
    formatted as a GitHub-flavored markdown table.
    """
    if response.data:
        headers = ["type", "id"] + list(
            response.data[0]
            .attributes.dict(exclude={"router", "service", "middlewares"})
            .keys()
        )
        tabular = []
        for resource in response.data:
            tabular.append(
                [resource.type, resource.id]
                + list(
                    resource.attributes.dict(
                        exclude={"router", "service", "middlewares"}
                    ).values()
                )
            )
        table = tabulate(tabular, headers=headers, tablefmt="github")
        return table



from typing import Generic, TypeVar, Optional, List, Dict
from pydantic import BaseModel
from pydantic.generics import GenericModel
from pathlib import Path


class RouterModel(BaseModel):
    """Models proxy router configs attached to Proximatic() provider resources."""

    id: str
    entryPoints: List[str] = ["web-secure"]
    rule: str
    middlewares: List[str] = []
    tls: dict = {}
    service: str


class LoadBalancerModel(BaseModel):
    """Models proxy loadBalancer settings inside of proxy service configs."""

    servers: List[dict] = [{"url": ""}]
    passHostHeader: bool = False


class ServiceModel(BaseModel):
    """Models proxy service configs attached to Proximatic() provider resources."""

    loadBalancer: LoadBalancerModel


class ProviderAttributesModel(BaseModel):
    """Models the data attributes of a Proximatic() provider resource."""

    router: RouterModel
    service: ServiceModel
    middlewares: dict
    endpoint: str
    server: str


AttributesModelType = TypeVar("AttributesModelType")

class ResourceModel(GenericModel, Generic[AttributesModelType]):
    """Models any resource in the Proximatic().system resources stores.
    The attributes parameter accepts any pydantic model, allowing flexible
    data schemas depending on the resource type."""

    id: str
    type: str
    attributes: AttributesModelType  # an attributes object representing some of the resource’s data.
    meta: Dict[str, str] = {}


class ResponseErrorModel(BaseModel):
    """Models all errors attached to responses generated by Proximatic()."""

    id: str  # a unique identifier for this particular occurrence of the problem.
    status: str = ""  # the HTTP status code applicable to this problem, expressed as a string value.
    code: str = ""  # an application-specific error code, expressed as a string value.
    title: str = ""  # a short, human-readable summary of the problem that SHOULD NOT change from occurrence to occurrence of the problem.
    detail: str = (
        ""  # a human-readable explanation specific to this occurrence of the problem.
    )
    meta: Dict[
        str, str
    ] = {}  # a meta object containing non-standard meta-information about the error.


class ResponseModel(BaseModel):
    """Models all responses generated by Proximatic()."""

    data: Optional[List[ResourceModel]]
    error: Optional[List[ResponseErrorModel]]
    meta: Optional[Dict[str, str]]


class SystemConfigModel(BaseModel):
    """Models the entire Proximatic().system configuration store."""

    yml_path: Path
    fqdn: str = "example.com"
    providers: List[ResourceModel] = []


class ProviderExportModel(BaseModel):
    """
    Models the exported YAML for provider resource file dumps.
    At export time, models are inserted into the top-level routers, services,
    and middlewares sections of the 'http' dictionary.
    """

    http: Dict[str, dict] = {"routers": {}, "services": {}, "middlewares": {}}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['proximatic']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'requests>=2.25.1,<3.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['proximatic = proximatic.main:cli',
                     'proximatic-http = proximatic.runhttp:http']}

setup_kwargs = {
    'name': 'proximatic',
    'version': '0.1.6',
    'description': 'Python API and CLI for managing Proximatic configuration.',
    'long_description': "# Proximatic\n\nPython API for managing Proximatic configuration files.\n\nWhen installed, the `proximatic` command provides a CLI for managing Proximatic configuration.\n\nThis Python package provides the core for the Proximatic system.\n\n## Installation\n\n```bash\npip install proximatic\n```\n\n## Usage\n\n### Command Line Interface (CLI)\n\nOpen a Terminal and type:\n\n```bash\nproximatic\n```\n\nUse `proximatic --help` for available commands and options.\n\n### Python API programmatic interface\n\n```python\nfrom proximatic import Proximatic\nproximatic = Proximatic(yml_dir='/path/to/your/proximatic/data', fqdn='yourdomain.org')\n```\n\n### JSON REST API interface\n\nYou can run the (experimental) REST API on your localhost by typing the command:\n\n```bash\nproximatic-http\n```\n\nIt will try to open your browser to [http://localhost:8000](http://localhost:8000)\n\n## License\n\nThe MIT License (MIT)\n\n## Author\n\nLink Swanson (LunkRat)",
    'author': 'Link Swanson',
    'author_email': 'link@swanson.link',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
